#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "acidblood.h"
#include "acidfuncs.h"
#include "internal.h"
#include "network.h"
#include "extern.h"
#include "tracking.h"
#include "startup.h"
#include "list.h"

struct servercstruct {
	struct servercstruct *prev ;
	char *name ;
	int port ;
	char *password ;
	int  conatt ;
	time_t nexttry ;
	time_t pingtimeout ;
	int reap ;
	struct socket_info *socket ;
	struct server_list *server ;
	struct servercstruct *next ;	
} ;

static struct servercstruct *serverclist ;
 
void init_serverinfo() 
{
	serverclist=NULL ;
}

static void server_action(int action, struct socket_info *socket)
{
	if (action == NET_READ) {
		if (getline(socket) == 1)
			parse_server(socket);
		return;
	}

	if (action == NET_WRITE) {
		if (IsConnecting(socket)) {
		        NotConnecting(socket);
		        SetConnected(socket);
		        NotWriteable(socket) ;
		                                                                
			register_to_server(socket);
		}
		return;
	}
		
}

void complete_connect(struct socket_info *socket)
{
	struct server_list *server ;
	
	SetRegistered(socket) ;
	
	/* if an away message exists, set away */
	if (botinfo->awaymsg != NULL) {
		irc_sprintf(socket, "AWAY:%s \n", botinfo->awaymsg);
	}
	
	server=GetServerBySocket(socket) ;
	
	/* send mode information */

	irc_sprintf(socket,"MODE %s %s \n", server->nick, botinfo->mode);
	if (botinfo->ns == 1) {
	} else {
		join_channels(socket);
	}

}

int register_to_server(struct socket_info *sinfo)
{
	struct server_list *server ;
	server=GetServerBySocket(sinfo) ;
	
	if (botinfo->srvpass) {
		irc_sprintf(sinfo, "PASS %s \n", botinfo->srvpass);
	}

	free(server->nick);
	server->nick = strdup(botinfo->nick) ; 

	/* user username hostname(ignored) servername(ignored) :realname */
	irc_sprintf(sinfo,"USER %s %s %s :%s \n",
		    botinfo->user, botinfo->server,
		    botinfo->mode, botinfo->fname);
	irc_sprintf(sinfo,"NICK %s \n", server->nick);
	return (0);
}

void AddServerConf(char *name, int port, char *password) {
	struct servercstruct *slist, *stmp ;
	stmp=malloc(sizeof(struct servercstruct)) ;
	
	
	LinkToList(stmp,slist,serverclist) ;

	slist->name=strdup(name) ;
	if(password!=NULL)
		slist->password=strdup(password) ;
	
	slist->port=port ;
	
	slist->socket=GetNewSocket(server_action) ;
	
	SetIrcProtocol(slist->socket) ;
	slist->nexttry=0  ;
	slist->pingtimeout=0 ;
	                
	if (slist->socket!=NULL)
		return ;

	UnlinkListItem(slist,serverclist) ;
	free(slist->name) ;
	free(slist->password) ;
	free(slist) ;		

}

static int CalcConDelay(int attempt) {
	/* no connects less than 10 seconds appart */
	if(attempt <= 3)
		return(10) ;
		
	/* exponentially increase the delay to avoid flooding servers*/
	return(attempt*attempt) ;		
}

static void CheckMissing(struct servercstruct *server) {
	int delay ;
	
	if(!IsConnecting(server->socket)&&(server->nexttry<=currtime) ) {
		if(server->server!=NULL) {
			DelServer(server->server) ;
			server->server=NULL ;
		}
		if(botinfo->maxtries != 0 && server->conatt >= botinfo->maxtries) {
			NotConnectable(server->socket) ;
			return ;
		}
			
		server->conatt++ ;
		delay=CalcConDelay(server->conatt) ;
		server->nexttry=currtime+delay;
		info_log("Connecting to %s attempt %i next attempt in %i seconds\n",server->name,server->conatt,delay) ; 
		AddServer(server->socket, server->name, botinfo->nick);
		connect_to_host(server->socket,server->name,server->port) ;
	}
	
}

static void ReapConfServers() {
	
	struct servercstruct *slist, *tnext ;
	
	slist=serverclist ;
	
	while(slist!=NULL) {
		tnext=slist->next ;
		
		if(slist->reap) {		
			UnlinkListItem(slist,serverclist) ;
			FreeSocket(slist->socket) ;
			free(slist->name) ;
			free(slist->password) ;
			free(slist) ;		
		}	
		slist=tnext ;
	}

}

void CheckConfServers() {
	struct servercstruct *slist ;
	
	ReapConfServers() ;
	
	slist=serverclist ;
	while(slist!=NULL) {
		if(!IsConnected(slist->socket)) { 
			CheckMissing(slist) ;
		} 
		
		if(IsRegistered(slist->socket) && slist->conatt != 0) {
			slist->conatt = 0 ;
			slist->nexttry = 0 ; 
			slist->pingtimeout = 0 ;
		}
		
		if(IsNotConnectable(slist->socket))
			slist->reap=1 ;
		
		slist=slist->next ;
	}
}

int parse_server (struct socket_info *sinfo)
{
	struct inputstruct is ;
	
	SplitInput(sinfo->buff,&is) ;
	
	if(isalpha((int)is.command[1])) {
		return(handle_server_command(sinfo,&is));
	}
            
	return(handle_numeric(sinfo,&is)) ;
              
	return(0) ;
}
